import getopt, sys, os, re
from Bio import SeqIO
from Bio.Blast import NCBIWWW, NCBIXML, NCBIStandalone
from Bio import Entrez



# global variables
dbpath = '/home/jleigh/Databases'
blastexe = '/usr/bin/blastall'
EvalRange = 1e10


def universalCode():
  """Return a dictionary of codon: aacode for universal genetic code."""
  
  code = {}
  
  code['TTT'] = code['TTC'] = 'F'
  code['TTA'] = code['TTG'] = 'L'
  code['TCT'] = code['TCC'] = code['TCA'] = code['TCG'] = 'S'
  code['TAT'] = code['TAC'] = 'Y'
  code['TAA'] = code['TAG'] = '*'
  code['TGT'] = code['TGC'] = 'C'
  code['TGA'] = '*'
  code['TGG'] = 'W'
  code['CTT'] = code['CTC'] = code['CTA'] = code['CTG'] = 'L'
  code['CCT'] = code['CCC'] = code['CCA'] = code['CCG'] = 'P'
  code['CAT'] = code['CAC'] = 'H'
  code['CAA'] = code['CAG'] = 'Q'
  code['CGT'] = code['CGC'] = code['CGA'] = code['CGG'] = 'R'
  code['ATT'] = code['ATC'] = code['ATA'] = 'I'
  code['ATG'] = 'M'
  code['ACT'] = code['ACC'] = code['ACA'] = code['ACG'] = 'T'
  code['AAT'] = code['AAC'] = 'N'
  code['AAA'] = code['AAG'] = 'K'
  code['AGT'] = code['AGC'] = 'S'
  code['AGA'] = code['AGG'] = 'R'
  code['GTT'] = code['GTC'] = code['GTA'] = code['GTG'] = 'V'
  code['GCT'] = code['GCC'] = code['GCA'] = code['GCG'] = 'A'
  code['GAT'] = code['GAC'] = 'D'
  code['GAA'] = code['GAG'] = 'E'
  code['GGT'] = code['GGC'] = code['GGA'] = code['GGG'] = 'G'
  
  
  # ambiguous codes
  code['TTY'] = 'F'
  code['TTR'] = code['YTR'] = 'L'
  code['TCY'] = code['TCR'] = code['TCW'] = code['TCS'] = code['TCM'] = \
  code['TCK'] = code['TCH'] = code['TCB'] = code['TCV'] = code['TCD'] = \
  code['TCN'] = 'S'
  code['TAY'] = 'Y'
  code['TAR'] = code['TRA'] = '*'
  code['TGY'] = 'C'
  code['CTY'] = code['CTR'] = code['CTW'] = code['CTS'] = code['CTM'] = \
  code['CTK'] = code['CTH'] = code['CTB'] = code['CTV'] = code['CTD'] = \
  code['CTN'] = 'L'
  code['CCY'] = code['CCR'] = code['CCW'] = code['CCS'] = code['CCM'] = \
  code['CCK'] = code['CCH'] = code['CCB'] = code['CCV'] = code['CCD'] = \
  code['CCN'] = 'P'
  code['CAY'] = 'H'
  code['CAR'] = 'Q'
  code['CGY'] = code['CGR'] = code['CGW'] = code['CGS'] = code['CGM'] = \
  code['CGK'] = code['CGH'] = code['CGB'] = code['CGV'] = code['CGD'] = \
  code['CGN'] = 'R'
  code['ATY'] = code['ATM'] = code['ATW'] = code['ATH'] = 'I'
  code['ACY'] = code['ACR'] = code['ACW'] = code['ACS'] = code['ACM'] = \
  code['ACK'] = code['ACH'] = code['ACB'] = code['ACV'] = code['ACD'] = \
  code['ACN'] = 'T'
  code['AAY'] = 'N'
  code['AAR'] = 'K'
  code['AGY'] = 'S'
  code['AGR'] = code['MGR'] = 'R'
  code['GTY'] = code['GTR'] = code['GTW'] = code['GTS'] = code['GTM'] = \
  code['GTK'] = code['GTH'] = code['GTB'] = code['GTV'] = code['GTD'] = \
  code['GTN'] = 'V'
  code['GCY'] = code['GCR'] = code['GCW'] = code['GCS'] = code['GCM'] = \
  code['GCK'] = code['GCH'] = code['GCB'] = code['GCV'] = code['GCD'] = \
  code['GCN'] = 'A'
  code['GAY'] = 'D'
  code['GAR'] = 'E'
  code['GGY'] = code['GGR'] = code['GGW'] = code['GGS'] = code['GGM'] = \
  code['GGK'] = code['GGH'] = code['GGB'] = code['GGV'] = code['GGD'] = \
  code['GGN'] = 'G'
  
  return code
  
def altcode(gbNum):
  """Return an alternative genetic code dictionary, using Genbank ID numbers."""
  
  # should delete ambiguity codes when appropriate!
  
  code = universalCode()
  
  if gbNum == 1: # universal code
    pass
  
  elif gbNum == 2: # vertebrate mt code
    code['AGA'] = code['AGG'] = code['AGR'] = '*'
    code['ATA'] = code['ATR'] = 'M'
    code['TGA'] = code['TGR'] = 'W'
  
  elif gbNum == 3: # yeast mitochondrial code
    # CGA and CGG should be absent, but I've left them coding for R (don't trust GB)
    code['ATA'] = code['ATR'] = 'M'
    code['CTT'] = code['CTC'] = code['CTA'] = code['CTG'] = 'T'
    code['CTY'] = code['CTR'] = code['CTW'] = code['CTS'] = code['CTM'] = \
    code['CTK'] = code['CTH'] = code['CTB'] = code['CTV'] = code['CTD'] = \
    code['CTN'] = 'T'
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 4: # mould mt et al. code
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 5: # invertebrate mitochondrial code
    code['AGA'] = code['AGG'] = code['AGY'] = code['AGR'] = code['AGW'] = \
    code['AGS'] = code['AGM'] = code['AGK'] = code['AGH'] = code['AGB'] = \
    code['AGV'] = code['AGD'] = code['AGN'] = 'S'

    code['ATA'] = code['ATR'] = 'M'
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 6: # ciliate code
    code['TAA'] = code['TAG'] = code['TAR'] = code['YAA'] = code['YAG'] = \
    code['YAR'] = 'Q'
    
  elif gbNum == 9: # flatworm and echinoderm mt code
    code['AAA'] = code['AAM'] = code['AAW'] = code['AAH'] = 'N' 
    code['AGA'] = code['AGG'] = code['AGY'] = code['AGR'] = code['AGW'] = \
    code['AGS'] = code['AGM'] = code['AGK'] = code['AGH'] = code['AGB'] = \
    code['AGV'] = code['AGD'] = code['AGN'] = 'S'
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 10: # euplotid code
    code['TGA'] = code['TGW'] = code['TGM'] = code['TGH'] = 'C'
    
  elif gbNum == 11: # bact. and platid code, only start codons are different
    pass
    
  elif gbNum == 12: # alt. yeast code
    code['CTG'] = 'S'
    
  elif gbNum == 13: # ascidian mt code
    code['AGA'] = code['AGG'] = code['AGR'] = code['RGG'] = code['RGA'] = \
    code['RGR'] = 'G'
    code['ATA'] = code['ATR'] = 'M'
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 14: # alt. flatworm mt code
    code['AAA'] = code['AAW'] = code['AAM'] = code['AAH'] = 'N'
    code['AGA'] = code['AGG'] = code['AGY'] = code['AGR'] = code['AGW'] = \
    code['AGS'] = code['AGM'] = code['AGK'] = code['AGH'] = code['AGB'] = \
    code['AGV'] = code['AGD'] = code['AGN'] = 'S' 
    code['TAA'] = code['TAW'] = code['TAM'] = code['TAH'] = 'Y'
    code['TGA'] = code['TGR'] = 'W'
    
  elif gbNum == 15: # blepharisma code
    code ['TAG'] = code['YAG'] = 'Q'
    
  elif gbNum == 16: # green algae mt code
    code['TAG'] = 'L'
    
  elif gbNum == 21: # trematode mt code
    code['TGA'] = code['TGR'] = 'W'
    code['ATA'] = code['ATY'] = 'M'
    code['AGA'] = code['AGG'] = code['AGY'] = code['AGR'] = code['AGW'] = \
    code['AGS'] = code['AGM'] = code['AGK'] = code['AGH'] = code['AGB'] = \
    code['AGV'] = code['AGD'] = code['AGN'] = 'S'
    
  elif gbNum == 22: # Scenedesmus obliquus mt
    code['TCA'] = code['TMA'] = code['TSA'] = code['TVA'] = '*'
    code['TAG'] = 'L'
    
  elif gbNum == 23: # Thraustochytrium aureum mt
    pass
    
  else:
    sys.stderr.write('Unknown genetic code: %d. Please see NCBI for valid codes.\n' % gbNum)
    sys.exit(0)
  
  return code

def translate(seq, frame, code):

  protseq = []
  
  for i in xrange(frame - 1, len(seq), 3):
    try:
      protseq.append(code[seq[i:i + 3]])
    except:
      protseq.append('X')
  
  return ''.join(protseq)
  
def flip(dnaseq):
  flipdict = {'A':'T', 'T':'A', 'G':'C', 'C':'G', 'N':'N'}
  
  # ambiguity codes:
  flipdict['Y'] = 'R'
  flipdict['R'] = 'Y'
  flipdict['W'] = 'W'
  flipdict['S'] = 'S'
  flipdict['M'] = 'K'
  flipdict['K'] = 'M'
  flipdict['H'] = 'D'
  flipdict['D'] = 'H'
  flipdict['B'] = 'V'
  flipdict['V'] = 'B'
  
  fliplist = [flipdict[char] for char in dnaseq]
  fliplist.reverse()
  
  return ''.join(fliplist)
 
def readTaxa(taxafilename, geneticCodes=None, skipgencode=False):
  tf = open(taxafilename)
  taxa = []
  
  for line in tf:
    if line.strip() == '':
      continue
    if not line.startswith(';'):
    
      if not skipgencode: #geneticCodes != None:
        try:
          line, code = line.rsplit(None, 1)
        except ValueError:
          print line
          raise
        #geneticCodes[line.strip().replace(' ', '_')] = altcode(int(code))
        geneticCodes[line.strip()] = altcode(int(code))
      #taxa.append(line.strip().replace(' ', '_'))
      taxa.append(line.strip())
      
      #################################################################
      ############# do something about space vs. _!!!!!! ##############
      #################################################################

  return taxa
  
def doNCBIblast(geneName, query, taxon, genCode, outfile, infofile, maxEval, EvalRange, maxattempts):
  for i in xrange(maxattempts):
    if i > 0:
      sys.stderr.write('Failed attempt %d, trying again.\n' % (i - 1))
      
    try:
      result = NCBIWWW.qblast('blastp', 'nr', query.seq.tostring(), entrez_query=taxon)
      
      if genCode:
        bestHits = extractBestHitsTranslate(result, maxEval, EvalRange, taxon, genCode)
      else:
        bestHits = extractBestHits(result, maxEval, EvalRange, taxon)
      break
    except ValueError:
      if i == (maxattempts - 1):
        sys.stderr.write('Failed final attempt. Exiting.\n')
        raise
        
  if len(bestHits) == 0:
    infofile.write('%s: no hits found!\n' % geneName)
  else:
    infofile.write('%s: best expect value: %g\n' % (geneName, bestHits[0]['expect']))
    infofile.write('%s\n' % bestHits[0]['title'])
    outfile.write('>%s %s ;\n%s\n' % (taxon, geneName, bestHits[0]['seq']))
        
    for i in xrange(1, len(bestHits)):
      infofile.write('Nearly as good, expect value: %g\n' % bestHits[i]['expect'])
      infofile.write('%s\n' % bestHits[i]['title'])
        
  infofile.write('\n')
  
def extractBestHits(blasthandle, expectCutoff, expectRange, taxonName):

  results = []
  giregex = re.compile(r'gi\|(\d+)\|')

  blastResults = NCBIXML.parse(blasthandle)
  bestEval = None
  for record in blastResults:
    for align in record.alignments:
      if taxonName not in align.title:
        continue

      Eval = float(align.hsps[0].expect)
      if Eval <= expectCutoff:
        if bestEval == None:
          bestEval = Eval
          try:
            ginum = giregex.search(align.title).group(1)
          except AttributeError:
            #infofile.write('Error: No GI number found in title %s.\n' % align.title)
            return None
          
          results.append({'title': align.title, 'seq': retrieveSeq(ginum), 'expect': Eval})

        else:
          if Eval > (bestEval * expectRange):
            break
          else:
            results.append({'title': align.title, 'seq': retrieveSeq(ginum), 'expect': Eval})

  return results
  
def retrieveSeq(seqID):
  handle = Entrez.efetch(db="protein", id=seqID, rettype="fasta")
  seq = handle.readlines()

  return ''.join(seq[1:]).replace('\n', '')

def extractBestHitsTranslate(blasthandle, expectCutoff, expectRange, taxonName, taxonCode):
  #print 'taxon:', taxonName

  results = []
  giregex = re.compile(r'gi\|(\d+)\|')

  blastResults = NCBIXML.parse(blasthandle)
  bestEval = None
  for record in blastResults:
    for align in record.alignments:
      if taxonName not in align.title:
        continue

      Eval = float(align.hsps[0].expect)
      if Eval <= expectCutoff:
        try:
          ginum = giregex.search(align.title).group(1)
        except AttributeError:
          return None
        if bestEval == None:
          bestEval = Eval
          resSeq = retrieveSeqTranslate(ginum, align.hsps[0].sbjct, align.hsps[0].frame[1], taxonCode)
          if resSeq:
            results.append({'title': align.title, 'seq': resSeq, 'expect': Eval})

        else:
          if Eval > (bestEval * expectRange):
            break
          else:
            resSeq = retrieveSeqTranslate(ginum, align.hsps[0].sbjct, align.hsps[0].frame[1], taxonCode)
            if resSeq:
              results.append({'title': align.title, 'seq': resSeq, 'expect': Eval})

  return results
  
def retrieveSeqTranslate(seqID, partSeq, frame, code):
  handle = Entrez.efetch(db="nucleotide", id=seqID, rettype="fasta")
  dnaseq = handle.readlines()
  dnaseq = ''.join(dnaseq[1:]).replace('\n', '')
  if len(partSeq) > 50:  
    partSeq = partSeq.replace('-', '')[10:-10]
  else:
    partSeq = partSeq.replace('-', '')[1:-1]
  
  
  if frame < 0:
    frame *= -1
    dnaseq = flip(dnaseq)
    
  protseq = translate(dnaseq, frame, code)  
    
  startidx = protseq.find(partSeq)
  endidx = startidx + len(partSeq)
  

  while startidx > 0:
    if protseq[startidx - 1] == '*':
      break
    startidx -= 1
    
  while endidx < len(protseq):
    if protseq[endidx] == '*':
      break
    endidx += 1

  if len(protseq[startidx:endidx]) > 10:
    return protseq[startidx:endidx]
    
  else:
    print 'Returning none. Frame:', frame
    print 'Sbjct:', partSeq
    print 'Hit (gi = %s):' % seqID, protseq

  return None
  
def doLocalBlast(geneName, query, taxa, genCodes, dbname, maxEval, EvalRange):
  blastdb = '%s/%s' % (dbpath, dbname)
  blastin = '.temp.fas'
  
  tempfile = open(blastin, 'w')
  tempfile.write('>%s\n%s' % (query.description, query.seq.tostring()))
  tempfile.close()
  
  result, error = NCBIStandalone.blastall(blastexe, 'tblastn', blastdb, blastin, expectation=str(maxEval))
  bestHits = extractBestHitsLocal(result, maxEval, EvalRange, taxa, genCodes, blastdb)
  
  if len(bestHits) == 0:
    sys.stderr.write('WARNING: No hits found. Did you run formatdb with \'-o T\'?\n')
    
  os.remove('.temp.fas')
  
  return bestHits
  
def recordHitsLocal(taxonHits, taxon, geneName, outfile, infofile):
  if len(taxonhits) == 0:
    infofile.write('%s: no hits found!\n' % geneName)
  else:
    infofiles.write('%s: best expect value: %g\n' % (geneName, taxonhits[0]['expect']))
    infofiles.write('%s\n' % taxonhits[0]['title'])
    outfile.write('>%s %s ;\n%s\n' % (taxon, geneName, taxonhits[0]['seq']))
        
  for i in xrange(1, len(taxonhits)):
    infofiles.write('Nearly as good, expect value: %g\n' % taxonhits[i]['expect'])
    infofiles.write('%s\n' % taxonhits[i]['title'])
        
  infofiles.write('\n')
  
def extractBestHitsLocal(blasthandle, expectCutoff, expectRange, taxa, geneticCodes, db):

  results = {}
  topEvals = {}
  
  for t in taxa:
    results[t] = []
      
  blastResults = NCBIXML.parse(blasthandle)
    
  for record in blastResults:
    for align in record.alignments:
      defline = align.title.split()[0]
      name = defline.split('|')[1].replace('_', ' ')
      hsp = align.hsps[0]
      
      try:
        taxonResults = results[name]
      except KeyError:
        continue
        
      Eval = float(align.hsps[0].expect)
      if Eval <= expectCutoff:

        if len(taxonResults) == 0:
          taxonResults.append({'title': align.title, 'seq': retrieveLocalSeq(defline, db, geneticCodes[name], hsp.sbjct, hsp.frame[0]), 'expect': Eval})
          topEvals[name] = Eval

        else:
          if Eval <= (topEvals[name] * expectRange):
            taxonResults.append({'title': align.title, 'seq': retrieveLocalSeq(defline, db, geneticCodes[name], hsp.sbjct, hsp.frame[0]), 'expect': Eval})
  
  return results


def retrieveLocalSeq(seqID, database, geneticCode, partSeq, frame):

  fastacmdbin = '/usr/bin/fastacmd'

  result = os.popen('%s -s "%s" -d %s' % (fastacmdbin, seqID, database))
  dnaseq = [line.strip() for line in result.readlines()[1:]]
  result.close()
  
  partSeq = partSeq.replace('-', '')
  
  dnaseq = ''.join(dnaseq)
  if frame < 0:
    frame *= -1
    dnaseq = flip(dnaseq)
  
  # Local BLAST uses universal code, need to use u. code temporarily
  protseq = translate(dnaseq, frame, ucode)
    
  startidx = protseq.find(partSeq)
  endidx = startidx + len(partSeq)
  
  # get real sequence before finding ends
  protseq = translate(dnaseq, frame, geneticCode)
  
  while startidx > 0:
    if protseq[startidx - 1] == '*':
      break
    startidx -= 1
    
  while endidx < len(protseq):
    if protseq[endidx] == '*':
      break
    endidx += 1

  #print 'Final:', protseq[startidx:endidx]
  return protseq[startidx:endidx]
 
def recordHits(hits, taxon, geneName, outfile, infofile):
  if len(hits) == 0:
    infofile.write('%s: no hits found!\n' % geneName)
  else:
    infofile.write('%s: best expect value: %g\n' % (geneName, hits[0]['expect']))
    infofile.write('%s\n' % hits[0]['title'])
    outfile.write('>%s %s ;\n%s\n' % (t, geneName, hits[0]['seq']))
        
  for i in xrange(1, len(hits)):
    infofile.write('Nearly as good, expect value: %g\n' % hits[i]['expect'])
    infofile.write('%s\n' % hits[i]['title'])
        
  infofile.write('\n')
  
def dogetopt(argv, reqdopts, legalopts):
  """Return a dictionary of options, check to make sure reqdopts are there."""
  
  if '-h' in argv:
    usage()
  
  opts, args = getopt.getopt(argv, legalopts)
  
  optdict = {}
  
  for (opt, val) in opts:
    optdict[opt[1:]] = val
   
  for opt in reqdopts:
    if opt not in optdict:
      raise getopt.GetoptError('Missing required option!')
      
  return (optdict, args)
  
  
def usage():
  sys.stderr.write('Usage: python blastmonkey.py [options]\n\n')
  sys.stderr.write('Required options:\n')
  sys.stderr.write('-t taxonfile  List of taxon names to BLAST against\n')
  sys.stderr.write('-q queryfile  FASTA-format file with query sequences\n')
  sys.stderr.write('-d database   Database to BLAST against (use nr or est to BLAST against NCBI)\n')
  sys.stderr.write('-p program    BLAST program (tblastn or blastp)\n')
  
  sys.stderr.write('\nOptional:\n')
  sys.stderr.write('-e eval       Expect value maximum cutoff (default = 10E-4)\n')
  sys.stderr.write('-n            Don\'t look for genetic code info in taxon name file\n')
  sys.stderr.write('-h            Print these instructions, then exit\n\n')
  
  sys.exit(1)
  
if __name__ == '__main__':

  genCodes = {}
  ucode = universalCode()

  try:
    optdict, args = dogetopt(sys.argv[1:], 'tqdp', 't:q:d:p:nh')
  except getopt.GetoptError:
    usage()

  if 'h' in optdict:
    usage()

  try:
    infile = open(optdict['q'])
  except OSError:
    sys.stderr.write('Input file %s not found!!\n' % optdict['q'])
    sys.exit(1)

  taxa = readTaxa(optdict['t'], genCodes, skipgencode=('n' in optdict))
  if 'e' in optdict:
    maxEval = float(optdict['e'])
  else:
    maxEval = 1e-4
    
  dbname = optdict['d']
  if dbname == 'nr' or dbname == 'est':
    localblast = False
  else:
    localblast = True
    
  blastprog = optdict['p']
  if blastprog == 'tblastn' and 'n' in optdict:
    sys.stderr.write('Error! You MUST include genetic codes to use tblastn!\n')
    sys.exit(0)
    
  
  infofiles = {}
  for t in taxa:
    infofiles[t] = open('%s.info' % t.replace(' ', '_'), 'w')
    
  for fastaSeq in SeqIO.parse(infile, "fasta"):
    wordlist = fastaSeq.description.split(';')
    if len(wordlist) < 2:
      sys.stderr.write('Input file MUST have a semicolon (;) before the gene name.\n')
      sys.exit(1)
    geneName = wordlist[0].split()[-1]
    outfile = open('%s.fas' % geneName, 'w')

    if blastprog == 'tblastn':
      if dbname == 'est': # BLAST against NCBI's est DB, not locally
        doNCBIblast(geneName, fastaSeq, t, genCodes[t], outfile, infofiles[t], maxEval, EvalRange, maxattempts=5)

      else: # tblastn vs. local (nucleotide) database
        bestHits = doLocalBlast(geneName, fastaSeq, taxa, genCodes, dbname, maxEval, EvalRange) 
        for t in taxa:
          recordHits(bestHits[t], t, geneName, outfile, infofiles[t])
        outfile.close()
      
    elif blastprog == 'blastp':
      if localblast or dbname == 'est':
        sys.stderr.write('Sorry! BLASTP is only enabled for NCBI BLAST vs. nr database.\n')
        sys.exit(1)
      for t in taxa:
        doNCBIblast(geneName, fastaSeq, t, None, outfile, infofiles[t], maxEval, EvalRange, maxattempts=5)
        
      outfile.close()  
  
    else:
      sys.stderr.write('Error! Unknown BLAST program: %s.\n' % blastprog)
      usage()
    

  for t in taxa:
    infofiles[t].close()




    